import allensdk.model.biophysical.runner as single_cell
import logging, os, sys, traceback, subprocess, logging.config as lc
import shutil
from allensdk.model.biophys_sim.config import Config
from pkg_resources import resource_filename #@UnresolvedImport


class RunSimulate(object):
    _log = logging.getLogger('allensdk.model.biophysical.run_simulate')
    
    def __init__(self,
                 input_json, 
                 output_json):
        self.input_json = input_json
        self.output_json = output_json
        self.app_config = None
        self.manifest = None
        
    
    def load_manifest(self):
        self.app_config = Config().load(self.input_json)
        self.manifest = self.app_config.manifest
        fix_sections = ['passive', 'axon_morph,', 'conditions', 'fitting']
        self.app_config.fix_unary_sections(fix_sections)

     
    def nrnivmodl(self):
        RunSimulate._log.debug("nrnivmodl")
        
        subprocess.call(['nrnivmodl', './modfiles'])    
    
    def simulate(self):
        from allensdk.internal.api.queries.biophysical_module_reader import BiophysicalModuleReader
        
        self.load_manifest()
                
        try:
            stimulus_path = self.manifest.get_path('stimulus_path')
            RunSimulate._log.info("stimulus path: %s" % (stimulus_path))
        except:
            raise Exception('Could not read input stimulus path from input config.')
        
        try:
            out_path = self.manifest.get_path('output_path')
            RunSimulate._log.info("result NWB file: %s" % (out_path))
        except:
            raise Exception('Could not read output path from input config.')
        
        try:
            morphology_path = self.manifest.get_path('MORPHOLOGY')
            RunSimulate._log.info("morphology path: %s" % (morphology_path))
        except:
            raise Exception('Could not read morphology path from input config.')
                
        single_cell.run(self.app_config)
        
        lims_upload_config = BiophysicalModuleReader()
        lims_upload_config.read_json(self.manifest.get_path('neuronal_model_run_data'))
        lims_upload_config.update_well_known_file(out_path)
        lims_upload_config.set_workflow_state('passed')
        lims_upload_config.write_file(self.output_json)
    
    
def main(command, lims_strategy_json, lims_response_json):
    ''' Entry point for module.
        :param command: select behavior, nrnivmodl or simulate
        :type command: string    
        :param lims_strategy_json: path to json file output from lims.
        :type lims_strategy_json: string
        :param lims_response_json: path to json file returned to lims.
        :type lims_response_json: string
    '''
    rs = RunSimulate(lims_strategy_json,
                     lims_response_json)

    RunSimulate._log.debug("command: %s" % (command))
    RunSimulate._log.debug("lims strategy json: %s" % (lims_strategy_json))
    RunSimulate._log.debug("lims upload json: %s" % (lims_response_json))
        
    log_config = resource_filename('allensdk.model.biophysical.run_simulate',
                                    'logging.conf')
    lc.fileConfig(log_config)
    os.environ['LOG_CFG'] = log_config
    
    if 'nrnivmodl' == command:
        rs.nrnivmodl()
    else:
        rs.simulate()


if __name__ == '__main__':
    command, input_json, output_json = sys.argv[-3:]
    
    try:
        main(command, input_json, output_json)
        RunSimulate._log.debug("success")
    except Exception as e:
        RunSimulate._log.error(traceback.format_exc())   
        exit(1)



import pytest
from mock import MagicMock
from allensdk.api.queries.image_download_api import ImageDownloadApi


@pytest.fixture
def image_api():
    image_api = ImageDownloadApi()
    
    image_api.retrieve_file_over_http = \
        MagicMock(name='retrieve_file_over_http')
    
    return image_api
    
    
def test_api_doc_url_download_image_downsampled(image_api):
    '''
    Notes
    -----
    See: `Experimental Overview and Metadata `<http://help.brain-map.org/display/mouseconnectivity/API#API-ExperimentalOverviewandMetadata>_
    , link labeled 'Download image downsampled by factor of 6 using default thresholds'.
    '''
    path = '126862575.jpg'
    
    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     range=[0,932, 0,1279, 0,4095])
    
    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&range=0,932,0,1279,0,4095",
        path)


def test_api_doc_url_download_image_full_resolution(image_api):
    '''
    Notes
    -----
    See: `Experimental Overview and Metadata `<http://help.brain-map.org/display/mouseconnectivity/API#API-ExperimentalOverviewandMetadata>_
    , link labeled 'Download a region of interest at full resolution using default thresholds'.
    '''
    expected = 'http://api.brain-map.org/api/v2/section_image_download/126862575?left=19045&top=11684&width=1000&height=1000&range=0,932,0,1279,0,4095'
    path = '126862575.jpg'

    image_api.retrieve_file_over_http = \
        MagicMock(name='retrieve_file_over_http')
    
    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     left=19045,
                                     top=11684,
                                     width=1000,
                                     height=1000,
                                     range=[0,932, 0,1279, 0,4095])
    
    image_api.retrieve_file_over_http.assert_called_once_with(expected, path)



import numpy as np
import matplotlib.pyplot as plt
import scipy.sparse as sparse
from scipy.linalg import solve_banded
from scipy.stats import linregress as linregress


def get_diagonals_from_sparse(mat):
    ''' Returns a dictionary of diagonals keyed by offsets

    Parameters
    ----------
    mat: scipy.sparse matrix

    Returns
    -------
    dictionary: diagonals keyed by offsets
    '''

    mat_dia = mat.todia()  #make sure the matrix is in diagonal format

    offsets = mat_dia.offsets
    diagonals = mat_dia.data

    mat_dict = {}

    for i,o in enumerate(offsets):
        mat_dict[o] = diagonals[i]

    return mat_dict

def ab_from_diagonals(mat_dict):
    ''' Constructs value for scipy.linalg.solve_banded

    Parameters
    ----------
    mat_dict: dictionary of diagonals keyed by offsets

    Returns
    -------
    ab: value for scipy.linalg.solve_banded
    '''
    offsets = mat_dict.keys()
    l = -np.min(offsets)
    u = np.max(offsets)

    T = mat_dict[offsets[0]].shape[0]

    ab = np.zeros([l+u+1,T])

    for o in offsets:
        index = u-o
        ab[index]=mat_dict[o]

    return ab

def error_calc(F_M, F_N, F_C, r):

    er = np.sqrt(np.mean(np.square(F_C-(F_M-r*F_N))))/np.mean(F_M)

    return er

def error_calc_outlier(F_M, F_N, F_C, r):

    std_F_M = np.std(F_M)
    mean_F_M = np.mean(F_M)
    ind_outlier = np.where(F_M > mean_F_M + 2.*std_F_M)

    er = np.sqrt(np.mean(np.square(F_C[ind_outlier]-(F_M[ind_outlier]-r*F_N[ind_outlier]))))/np.mean(F_M[ind_outlier])

    return er


class NeuropilSubtract (object):

    def __init__(self,T,lam=0.1):

        self.r = 0.001  # initial guess for neuropil subtraction constant
        self.T = T

        Ls = -sparse.eye(T-1,T,format='csr') + sparse.eye(T-1,T,1,format='csr')  #using csr because multiplication is fast
        dt = 1.0
        Ls /= dt
        Ls2 = Ls.T.dot(Ls)

        self.Ls2 = Ls2
        
        self.lam = lam

        M = sparse.eye(T) + lam*Ls2
        mat_dict = get_diagonals_from_sparse(M)
        ab = ab_from_diagonals(mat_dict)

        self.M = M
        self.ab = ab
        
    def set_F(self,F_M,F_N,F_M_crossval,F_N_crossval):
        ''' Internal initializatin routine
        '''
        self.F_M = F_M
        self.F_N = F_N
        self.F_M_crossval = F_M_crossval
        self.F_N_crossval = F_N_crossval

    
    #def fit_grad_desc_early_stop(self,r_init=0.001,learning_rate=0.1):
    def fit_grad_descent(self,r_init=0.001,learning_rate=0.1):
        ''' Calculate fit using gradient decent, as opposed to iteratively
        computing exact solutions
        '''
        delta_r = 1
        r = r_init

        r_list = []
        error_list = []

        F_C = solve_banded((1,1),self.ab,self.F_M - r*self.F_N)

        itmax = 10000
        it = 0

        exceed_bounds = False
        while (delta_r > 0.0001 and it < itmax):

            F_C = solve_banded((1,1),self.ab,self.F_M - r*self.F_N)
            r_new = r + learning_rate*np.mean((self.F_M - F_C - r*self.F_N)*self.F_N)

            '''compute error on cross-validation set'''
            F_C_crossval = solve_banded((1,1),self.ab,self.F_M_crossval - r*self.F_N_crossval)
#            error_it = error_calc_outlier(self.F_M_crossval, self.F_N_crossval, F_C_crossval, r)
            error_it = abs(error_calc(self.F_M_crossval, self.F_N_crossval, F_C_crossval, r))
            
            delta_r = np.abs(r_new - r)/r
            r = r_new

            r_list.append(r)
            error_list.append(error_it)
            it+=1
            
            if error_it > error_list[it-1]: # early stopping
                break

        # if r or error_it go out of acceptable bounds, break 
        if r < 0.0 or r > 1.0:
            exceed_bounds = True
        if error_it > 0.2:
            exceed_bounds = True

        F_C = solve_banded((1,1),self.ab,self.F_M - r*self.F_N)

        r_list = np.array(r_list)
        error_list = np.array(error_list)
        self.r_vals = r_list
        self.error_vals = error_list
        self.r = self.r_vals[-1]
        self.F_C = F_C
        self.F_C_crossval = F_C_crossval
        self.it = it

        return exceed_bounds


# 'M' denotes ROI trace
# 'N' denotes neuropil trace
def estimate_contamination_ratios(F_M_unscaled, F_N_unscaled):
    ''' Calculates neuropil contamination of ROI

    Parameters
    ----------
    F_M_unscaled: ROI trace

    F_N_unscaled: Neuropil trace

    Returns
    -------
    dictionary: key-value pairs
        * 'r': the contamination ratio -- corrected trace = M - r*N
        * 'err': RMS error
        * 'min_error': minimum error
        * 'bounds_error': boolean. True if error or R are outside tolerance
    '''

    T = len(F_M_unscaled)
    assert T == len(F_N_unscaled), "Input arrays of different dimension"
    T_cross_val = int(T/2)
    if T - T_cross_val > T_cross_val:
        T = T - 1
    ns = NeuropilSubtract(T_cross_val, lam=0.05)
    F_M_unscaled_cross_val = np.copy(F_M_unscaled)
    F_N_unscaled_cross_val = np.copy(F_N_unscaled)

    '''pick r on first half of the trace '''
    F_M_unscaled = F_M_unscaled[0:T_cross_val]
    F_N_unscaled = F_N_unscaled[0:T_cross_val]        
    
    F_M_unscaled_cross_val = F_M_unscaled_cross_val[T_cross_val:T]
    F_N_unscaled_cross_val =  F_N_unscaled_cross_val[T_cross_val:T]
    
    ''' normalize to have F_N in (0,1)'''
    F_M = (F_M_unscaled - float(np.amin(F_N_unscaled)))/float(np.amax(F_N_unscaled)-np.amin(F_N_unscaled))
    F_N = (F_N_unscaled - float(np.amin(F_N_unscaled)))/float(np.amax(F_N_unscaled)-np.amin(F_N_unscaled))
    F_M_cross_val = (F_M_unscaled_cross_val - float(np.amin(F_N_unscaled_cross_val)))/float(np.amax(F_N_unscaled_cross_val)-np.amin(F_N_unscaled_cross_val))
    F_N_cross_val = (F_N_unscaled_cross_val - float(np.amin(F_N_unscaled_cross_val)))/float(np.amax(F_N_unscaled_cross_val)-np.amin(F_N_unscaled_cross_val))

    # fitting model 
    
    ns.set_F(F_M, F_N, F_M_cross_val, F_N_cross_val)

    '''stop gradient descent at first increase of cross-validation error'''
    bounds_err = ns.fit_grad_descent(learning_rate=10)
            
    F_C_unscaled = ns.F_C*float(np.amax(F_N_unscaled)-np.amin(F_N_unscaled)) + (1-ns.r)*float(np.amin(F_N_unscaled))
    F_C_unscaled_crossval = ns.F_C_crossval*float(np.amax(F_N_unscaled_cross_val)-np.amin(F_N_unscaled_cross_val)) + (1-ns.r)*float(np.amin(F_N_unscaled_cross_val))

    min_error = np.zeros(T)
    min_error[:T-T_cross_val] = F_C_unscaled
    min_error[T-T_cross_val:T] = F_C_unscaled_crossval        
    results = {}
    results["r"] = ns.r
    results["err"] = abs(ns.error_vals[-1])
    results["min_error"] = min_error
    results["bounds_error"] = bounds_err
    return results




# Copyright 2015 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from mock import patch, mock_open
from allensdk.config.model.description_parser import DescriptionParser


@pytest.fixture
def pyconfig():
    file_1 = ("{\n"
              "    \"section_A\": [\n"
              "       {\n"
              "           \"prop_a\": \"val_a\",\n"
              "           \"prop_b\": \"val_b\"\n"
              "       },\n"
              "       {\n"
              "           \"prop_c\": \"val_c\",\n"
              "           \"prop_d\": \"val_d\"\n"
              "       }\n"
              "    ],\n"
              "    \"section_B\": [\n"
              "       {\n"
              "           \"prop_e\": \"val_e\",\n"
              "           \"prop_f\": \"val_f\"\n"
              "       },\n"
              "       {\n"
              "           \"prop_g\": \"val_g\",\n"
              "           \"prop_h\": \"val_h\"\n"
              "       }\n"
              "    ]\n"
              "}\n"
              )
    file_2 = ("{\n"
              "    \"section_B\": [\n"
              "       {\n"
              "           \"prop_i\": \"val_i\",\n"
              "           \"prop_j\": \"val_j\"\n"
              "       }\n"
              "    ],\n"
              "    \"section_C\": [\n"
              "       {\n"
              "           \"prop_k\": \"val_k\",\n"
              "           \"prop_l\": \"val_l\"\n"
              "       }\n"
              "    ]\n"
              "}\n"
              )
    
    with patch("__builtin__.open",
               mock_open(
                   read_data=file_1)):
        parser = DescriptionParser()
        description = parser.read("mock_1.pycfg")
    
    with patch("__builtin__.open",
               mock_open(
                   read_data=file_2)):
        parser = DescriptionParser()
        parser.read("mock_2.pycfg",
                    description)
    
    return description


def testAllSectionsPresent(pyconfig):
    assert('section_A' in pyconfig.data and
           'section_B' in pyconfig.data and
           'section_C' in pyconfig.data)
    assert(len(pyconfig.data.keys()) == 3)


def testSectionA(pyconfig):
    assert len(pyconfig.data['section_A']) == 2
    assert pyconfig.data['section_A'][0] == {
        'prop_a': 'val_a',
        'prop_b': 'val_b' }
    assert pyconfig.data['section_A'][1] == {
        'prop_c': 'val_c',
        'prop_d': 'val_d' }


def testSectionB(pyconfig):
    assert len(pyconfig.data['section_B']) == 3
    assert pyconfig.data['section_B'][0] == {
        'prop_e': 'val_e',
        'prop_f': 'val_f' }
    assert pyconfig.data['section_B'][1] == {
        'prop_g': 'val_g',
        'prop_h': 'val_h' }
    assert pyconfig.data['section_B'][2] == {
        'prop_i': 'val_i',
        'prop_j': 'val_j' }


def testSectionC(pyconfig):
    assert len(pyconfig.data['section_C']) == 1
    assert pyconfig.data['section_C'][0] == {
        'prop_k': 'val_k',
        'prop_l': 'val_l' }
